/*
 * packet.C
 * (c)1997 pcg@goof.com
 */

#include <config.h>

#include <stdio.h>	// good ol' stdio

#include "packet.h"
#include "util.h"
#include "global.h"

enum ie_tag {
	BEARER_CAPABILITY		= 0x04,
	DISPLAY				= 0x28,
	CALLING_PARTY_NUMBER		= 0x6c,
	CALLING_PARTY_SUBADDRESS	= 0x6d,
	CALLED_PARTY_NUMBER		= 0x70,
	CALLED_PARTY_SUBADDRESS		= 0x71,
	LOW_LAYER_COMPATIBILITY		= 0x7c,
	HIGH_LAYER_COMPATIBILITY	= 0x7d
};	

packet::packet (const char *src, const char *dst, ie_mask pmask)
{
  mask = pmask;
  strcpy (src_msn, src);
  strcpy (dst_msn, dst);
}

packet::packet (const packet &p)
: vector(p)
{
  mask = p.mask;
  strcpy (src_msn, p.src_msn);
  strcpy (dst_msn, p.dst_msn);
  time = p.time;
}

const packet &packet::operator =(const packet &p)
{
  *(vector *)this = *(vector *)&p;
  mask = p.mask;
  strcpy (src_msn, p.src_msn);
  strcpy (dst_msn, p.dst_msn);
  time = p.time;
  return *this;
}

packet::packet (const char *raw)
{
  u8 *p, *e;
  u8 data[300]; // 260 octets max. size anyway
  int codeset = 0, maincodeset = 0;
  
  raw += 8; // skip "PACKET: "
  
  mask = IEM_NONE;
  src_msn[0] = 0;
  dst_msn[0] = 0;
  
  p = e = data;
  while (*raw >= '0')
    {
      int i;
      
      if (sscanf (raw, "%02x", &i) != 1)
        break;
      
      *e++ = i;
      raw += 2;
    }
  
  /* skip protocol discriminator, callref and message type */
  p += (p[1] & 0x0f) + 3;
  
  u8 bits = 0; // calm down gcc by initializing
  int bitcnt = 8;
#define putbit(b) ({ bits = (bits>>1) | ((b)&0x80); if (!--bitcnt) { put(bits); bitcnt = 8; } })
#define putbits(b,c) { u8 B=(b)<<(8-(c)); int C=(c); while(C--) putbit(B), B<<=1; }
  
  while (p < e) {
    if ((*p & 0xf0) == 0x90) { // shift codeset
      codeset = *p & 0x07;
      
      if (!(*p & 0x08))
        maincodeset = codeset;
    }
    if (*p & 0x80)
      p++;
    else {
      if (codeset == 0) {
        int l = p[1];
        switch (*p) {
          case BEARER_CAPABILITY:
            mask |= IEM_BC;
            putbits (p[5], 7);
            putbits (p[6], 7);
            putbits (p[7], 7);
            putbits (p[8], 7);
            break;
            
	  case CALLING_PARTY_NUMBER:
	    // check format /*D*/
            src_msn[0] = '0'; // prepend '0' (only in germany?) /*D*/
	    memcpy (src_msn+1, p+4, l-2); src_msn[l-1] = 0;
	    normalize_msn (src_msn);
	    break;
            
	  case CALLED_PARTY_NUMBER:
	    // check format /*D*/
	    memcpy (dst_msn, p+3, l-1); dst_msn[l-1] = 0;
	    normalize_msn (dst_msn);
            break;
            
	  case CALLING_PARTY_SUBADDRESS:
	    mask |= IEM_CIS;
	    putbits (p[2], 4);
	    { int count=20; u8 *i=p+3; while(count--) putbits(*i++,8); }
            break;
            
	  case CALLED_PARTY_SUBADDRESS:
	    mask |= IEM_CDS;
	    putbits (p[2], 4);
	    { int count=20; u8 *i=p+3; while(count--) putbits(*i++,8); }
            break;
            
	  case LOW_LAYER_COMPATIBILITY:
	    mask |= IEM_LLC;
	    { int count=14; u8 *i=p+2; while(count--) putbits(*i++,8); }
            break;
            
	  case HIGH_LAYER_COMPATIBILITY:
	    mask |= IEM_HLC;
	    putbits (p[2], 8);
	    putbits (p[3], 8);
	    putbits (p[4], 8);
            break;
          
          default:
            break;
        }
      }
      p += p[1] + 2;
      codeset = maincodeset;
    }
  }
}

packet::operator const char * () const
{
  static char buff[600];
  char *b;
  const_iterator data = begin ();
  
  int bits = 0;
  u8 byte = 0; // calm down gcc by initializing
  
#define put(c) b += sprintf (b, "%02x", int(c))
#define getbit ({ if (--bits <= 0 && data != end ()) { byte = *data++; bits = 8; }; \
                  int r = byte & 1; byte >>= 1; r; })
#define getbits(n) ({ int r = 0, c = n; while (c--) r = (r << 1) | getbit; r; })

  b = buff;
  
  put (0xa1); // setup complete
  
  assert (mask & IEM_BC); // *must* be present ;)
  
  /* bearer capability. */
  put (BEARER_CAPABILITY); // die ersten 5 bytes sind wichtig, damit der kernel das paket erkennt!
  put (0x07);
  put (0x88); // Unrestricted with tones/announcements.
  put (0x90);
  put (0x21); // 5
  put (getbits(7)); // 5a
  put (getbits(7)); // 5b
  put (getbits(7)); // 5c
  put (0x80 | getbits(7)); // 5d
  
  put (CALLING_PARTY_NUMBER);
  put (2 + strlen (src_msn));
  put (0x01);
  put (0x81);
  for (const char *src = src_msn; *src; src++)
    put (*src & 0x7f);
  
  if (mask & IEM_CIS && data != end ())
    {
      put (CALLING_PARTY_SUBADDRESS);
      put (21);
      put (0xa0 | getbits(4)); /* 1-3 potentially unsafe. */
      int count = 20;
      while(count--)
        put (getbits(8));
    }
  
  put (CALLED_PARTY_NUMBER); // called party number
  put (1 + strlen (dst_msn));
  put (0x81);
  for (const char *dst = dst_msn; *dst; dst++)
    put (*dst & 0x7f);
  
  if (mask & IEM_CDS && data != end ())
    {
      put (CALLED_PARTY_SUBADDRESS);
      put (21);
      put (0xa0 | getbits(4)); /* 1-3 potentially unsafe. */
      int count = 20;
      while(count--)
        put (getbits(8));
    }
  
  if (mask & IEM_LLC && data != end ())
    {
      put (LOW_LAYER_COMPATIBILITY);
      put (14); /* potentially unsafe */
      int count = 14;
      while(count--)
        put (getbits(8));
    }
  
  if (mask & IEM_HLC && data != end ())
    {
      put (HIGH_LAYER_COMPATIBILITY);
      put (3);
      put (getbits(8)); /* potentially unsafe */
      put (getbits(8));
      put (getbits(8));
    }
  
  *b++ = 0;
  
  return buff;
}

int packet_length (ie_mask mask)
{
  int r = 0;
  
  if (mask & IEM_BC)		r +=  28;
  if (mask & IEM_CIS)		r += 164;
  if (mask & IEM_CDS)		r += 164;
  if (mask & IEM_LLC)		r += 112;
  if (mask & IEM_HLC)		r +=  24;
  
  return r>>3;
}

void normalize_msn (char *number)
{
  msn old;
  
  strcpy (old, number);
  
  if (old[0] != '0')
      strcat (strcat (strcpy (number, "0"), cfg_area), old);
/*  if (old[0] != '0')
    strcat (strcat (strcat (strcpy (number, "00"), cfg_country), cfg_area), old);
  else
    if (old[1] != '0')
      strcat (strcat (strcpy (number, "00"), cfg_country), &old[1]);*/
}

