#ifndef THREADPP_H
#define THREADPP_H

#include <pthread.h>

// needed because in bugs in libpthreads
__BEGIN_DECLS
#include <semaphore.h>
__END_DECLS

// for debugging
#include <stdio.h>

// mutex wrapper
class mutex {
  pthread_mutex_t m;
public:
  mutex(int recursive = 0);
#if 1
  void lock() { pthread_mutex_lock (&m); };
  void unlock() { pthread_mutex_unlock (&m); };
#else
  void lock() {
   fprintf (stderr, "%d: lock %08p\n", pthread_self (), this);
   pthread_mutex_lock (&m);
   fprintf (stderr, "%d: locked %08p\n", pthread_self (), this);
  };
  void unlock() {
   fprintf (stderr, "%d: unlock %08p\n", pthread_self (), this);
   pthread_mutex_unlock (&m);
   fprintf (stderr, "%d: unlocked %08p\n", pthread_self (), this);
  };
#endif
};

// guard mutex in this scope
struct guard {
  mutex &m;
  
  guard (mutex &mutex_to_lock) : m(mutex_to_lock) { m.lock (); };
  ~guard () { m.unlock (); };
};

// mutex wrapper
class recursive_mutex : public mutex {
  pthread_mutex_t m;
public:
  recursive_mutex() : mutex(1) {};
};

// semaphore wrapper
class semaphore {
  sem_t sem;
public:
  semaphore(int value = 0, bool shared = false) { sem_init (&sem, (int)shared, value); };
  ~semaphore() { sem_destroy (&sem); };
  void wait() { sem_wait (&sem); };
  bool trywait() { return sem_trywait (&sem) == 0; };
  void post() { sem_post (&sem); };
  int getvalue() { int sval; sem_getvalue (&sem, &sval); return sval; };
  int operator ()() { return getvalue (); };
};

struct runnable {
  typedef void *(runnable::*thread)(void *);
  int run (pthread_t *tid, thread t, void *arg = 0);
};

#endif

